

# Single Responsibility

# Open a stream between the computer and a camera?
# Open a connection to a usb attached webcam using OpenCV
# Feeds frames to someone else
class Camera:

    def __init__(self):
        pass

